#include <gtk/gtk.h>
#include <glib/gi18n.h>
#include <glib/gstdio.h>
#include <gio/gunixmounts.h>
#include <gio/gunixoutputstream.h>
#include <locale.h>
#include "gtk-compat.c" //requires libintl.h (locale.h)
#include "thunar-encryption.h"
#include "config.h"

//Static functions
static void thunarx_encryption_menu_provider_init (ThunarxMenuProviderIface *iface);
static GList *get_folder_actions (GtkWidget *window, ThunarxFileInfo *file);
static GList *thunarx_encryption_get_file_actions (ThunarxMenuProvider *provider, GtkWidget *window, GList *files);
static GList *thunarx_encryption_get_folder_actions (ThunarxMenuProvider *provider, GtkWidget *window, ThunarxFileInfo *folder);

static void thunarx_encryption_shell (ThunarxMenuItem *menuitem, gpointer *data);
static void thunarx_encryption_open (ThunarxMenuItem *menuitem, gpointer *data);
static void thunarx_encryption_close (ThunarxMenuItem *menuitem, gpointer *data);
static void close_child_watch_cb (GPid pid, gint status, gpointer user_data);
static void thunarx_encryption_encrypt_fuse (ThunarxMenuItem *menuitem, gpointer *data);
static void thunarx_encryption_encrypt_fs (ThunarxMenuItem *menuitem, gpointer *data);

static void do_ask_password (const gchar *path, FolderStatus status);
static void on_password_validate (GtkButton *button, GtkWidget *entry);
static void do_notify_open (const gchar *path, gboolean success);
static void do_notify_close (const gchar *path, gboolean success);

static GList* get_fscrypt_protectors (const gchar* path);

static void spawn_with_input_pipe_and_path (gchar **argv, gchar **envp, const gchar *input_string, const gchar *path);
static void open_child_watch_cb (GPid pid, gint status, gpointer user_data);



//GObject
struct _ThunarxEncryptionClass { GObjectClass __parent__; };
struct _ThunarxEncryption { GObject __parent__; };
THUNARX_DEFINE_TYPE_WITH_CODE (ThunarxEncryption, thunarx_encryption, G_TYPE_OBJECT,
    THUNARX_IMPLEMENT_INTERFACE (THUNARX_TYPE_MENU_PROVIDER, thunarx_encryption_menu_provider_init));
static void thunarx_encryption_class_init (ThunarxEncryptionClass *cls) {}
static void thunarx_encryption_init (ThunarxEncryption *encryption) {}



// Menus
static void thunarx_encryption_menu_provider_init (ThunarxMenuProviderIface *iface) {
    iface->get_file_menu_items = thunarx_encryption_get_file_actions;
    iface->get_folder_menu_items = thunarx_encryption_get_folder_actions;
}

static GList *thunarx_encryption_get_file_actions (ThunarxMenuProvider *provider, GtkWidget *window, GList *files) {
    if (G_LIKELY (files != NULL && files->next == NULL && thunarx_file_info_is_directory (files->data))) {
        return get_folder_actions (window, files->data);
    } else {
        return NULL;
    }
}

static GList* thunarx_encryption_get_folder_actions (ThunarxMenuProvider *provider, GtkWidget *window, ThunarxFileInfo *file) {
    return get_folder_actions (window, file); //inside folder
}

static GList *get_folder_actions (GtkWidget *window, ThunarxFileInfo *file) {
    setlocale (LC_ALL, "");
    bindtextdomain (PROJECT_NAME, LOCALE_DIR);
    bind_textdomain_codeset (PROJECT_NAME, "UTF-8");
    textdomain (PROJECT_NAME);

    GFile *folder = thunarx_file_info_get_location (file);
    GList *menu_items = NULL;

    ThunarxMenuItem *menu_shell = thunarx_menu_item_new ("ThunarxEncryption::shell", gettext ("Open terminal here"),
                                                         gettext ("Open a terminal in this folder"), "utilities-terminal");
    menu_items = g_list_prepend (menu_items, menu_shell);
    g_signal_connect (G_OBJECT (menu_shell), "activate", G_CALLBACK (thunarx_encryption_shell), window);
    g_object_set_data_full (G_OBJECT (menu_shell), "path", g_file_get_path (folder), g_free);

    gboolean has_fscrypt = is_fscrypt_available();
    gboolean has_cryfs = is_cryfs_available();
    gboolean has_encfs = is_encfs_available();

    FolderStatus status = get_folder_status (folder);
    if (status == ENCFS_FOLDER && has_encfs || status == CRYFS_FOLDER && has_cryfs || status == FSCRYPT_FOLDER && has_fscrypt) {
        ThunarxMenuItem *menu_open = thunarx_menu_item_new ("ThunarxEncryption::open", gettext ("Open encrypted folder"),
                                                            gettext ("Open (mount / unlock) encrypted folder"), "encrypted-folder");
        menu_items = g_list_prepend (menu_items, menu_open);
        g_signal_connect (G_OBJECT (menu_open), "activate", G_CALLBACK (thunarx_encryption_open), window);
        g_object_set_data_full (G_OBJECT (menu_open), "path", g_file_get_path (folder), g_free);
        g_object_set_data_full (G_OBJECT (menu_open), "status", g_strdup_printf ("%u", status), g_free);
    } else if (status == UNLOCKED_FOLDER) { //EncFS & CryFS already have umount menu in Thunar
        ThunarxMenuItem *menu_close = thunarx_menu_item_new ("ThunarxEncryption::close", gettext ("Close encrypted folder"),
                                                             gettext ("Close (lock) encrypted folder"), "encrypted-folder");
        menu_items = g_list_prepend (menu_items, menu_close);
        g_signal_connect (G_OBJECT (menu_close), "activate", G_CALLBACK (thunarx_encryption_close), window);
        g_object_set_data_full (G_OBJECT (menu_close), "path", g_file_get_path (folder), g_free);
    } else if (status == EMPTY_FOLDER) {
        if (has_cryfs || has_encfs) {
            ThunarxMenuItem *menu_encrypt_fuse = thunarx_menu_item_new ("ThunarxEncryption::crypt_fuse", gettext ("Encrypt folder (CryFS / EncFS)"),
                                                                        gettext ("Encrypt empty folder using CryFS or EncFS"), "encrypted-folder-new");
            menu_items = g_list_prepend (menu_items, menu_encrypt_fuse);
            g_signal_connect (G_OBJECT (menu_encrypt_fuse), "activate", G_CALLBACK (thunarx_encryption_encrypt_fuse), window);
            g_object_set_data_full (G_OBJECT (menu_encrypt_fuse), "path", g_file_get_path (folder), g_free);
        }
        if (has_fscrypt) {
            ThunarxMenuItem *menu_encrypt_fs = thunarx_menu_item_new ("ThunarxEncryption::crypt_fs", gettext ("Encrypt folder (FSCrypt)"),
                                                                      gettext ("Encrypt empty folder using FSCrypt"), "encrypted-folder-new");
            menu_items = g_list_prepend (menu_items, menu_encrypt_fs);
            g_signal_connect (G_OBJECT (menu_encrypt_fs), "activate", G_CALLBACK (thunarx_encryption_encrypt_fs), window);
            g_object_set_data_full (G_OBJECT (menu_encrypt_fs), "path", g_file_get_path (folder), g_free);
        }
    }

    g_object_unref (folder);
    return menu_items;
}



//Menus actions
static void thunarx_encryption_shell (ThunarxMenuItem *menuitem, gpointer *data) {
    const gchar *path = g_object_get_data (G_OBJECT (menuitem), "path");
    gchar *argv[] = { "exo-open", "--launch", "TerminalEmulator", "--working-directory", (gchar *) path, NULL };
    g_spawn_async_with_pipes (NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
}

static void thunarx_encryption_open (ThunarxMenuItem *menuitem, gpointer *data) {
    const gchar *path = g_object_get_data (G_OBJECT (menuitem), "path");
    FolderStatus status = (FolderStatus) g_ascii_strtoll (g_object_get_data (G_OBJECT (menuitem), "status"), NULL, 10);
    do_ask_password (path, status);
}

static void thunarx_encryption_close (ThunarxMenuItem *menuitem, gpointer *data)  { //only for FSCrypt
    const gchar *path = g_object_get_data (G_OBJECT (menuitem), "path");
    gchar *argv[] = { "fscrypt", "lock", (gchar *) path, NULL };
    GPid child_pid;
    if (g_spawn_async_with_pipes (NULL, argv, NULL, G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_STDOUT_TO_DEV_NULL,
                                  NULL, NULL, &child_pid, NULL, NULL, NULL, NULL)) {
        g_child_watch_add (child_pid, close_child_watch_cb, g_strdup (path)); //g_object_set_data_full -> g_free(path)
    }
}

static void close_child_watch_cb (GPid pid, gint status, gpointer user_data) {
    gchar *path = (gchar *) user_data;
    do_notify_close (path, g_spawn_check_wait_status (status, NULL) && !is_unlocked (path));
    g_free (path);
}

static void thunarx_encryption_encrypt_fuse (ThunarxMenuItem *menuitem, gpointer *data)  {
    const gchar *path = g_object_get_data (G_OBJECT (menuitem), "path");
    do_ask_password (path, DO_ENCRYPT_FUSE);
}

static void thunarx_encryption_encrypt_fs (ThunarxMenuItem *menuitem, gpointer *data)  {
    const gchar *path = g_object_get_data (G_OBJECT (menuitem), "path");
    do_ask_password (path, DO_ENCRYPT_FS);
}



//Helper functions
FolderStatus get_folder_status (GFile *folder) {
    gchar *path = g_file_get_path (folder);

    FolderStatus status = NORMAL_FOLDER;

    if (is_encfs (path)) status = ENCFS_FOLDER;
    else if (is_cryfs (path)) status = CRYFS_FOLDER;
    else if (is_fscrypt (path)) status = FSCRYPT_FOLDER;

    if (status == NORMAL_FOLDER) {
        if (is_empty (folder)) status = EMPTY_FOLDER;
    } else if (status == ENCFS_FOLDER || status == CRYFS_FOLDER) {
        if (is_mounted (path)) status = MOUNTED_FOLDER;
    } else { //FSCRYPT_FOLDER
        if (is_unlocked (path)) status = UNLOCKED_FOLDER;
    }

    g_free (path);
    return status;
}

gboolean is_mounted (const gchar *path) {
    gboolean mounted = FALSE;
    GVolumeMonitor *vm = g_volume_monitor_get ();
    GList *mounts = g_volume_monitor_get_mounts (vm);
    for (GList *current = mounts; !mounted && current != NULL; current = current->next) {
        gchar *_path = g_mount_get_name (current->data);
        if (g_strcmp0 (path, _path) == 0) mounted = TRUE;
        g_free (_path);
    }
    g_list_free_full (mounts, g_object_unref);
    g_object_unref (vm);
    return mounted;
}

gboolean is_unlocked (const gchar *path) {
    gchar *output = NULL;
    gchar *argv[] = { "fscrypt", "status", (gchar *) path, NULL };
    gboolean unlocked = FALSE;
    if (g_spawn_sync (NULL, argv, NULL, G_SPAWN_SEARCH_PATH + G_SPAWN_STDERR_TO_DEV_NULL, NULL, NULL, &output, NULL, NULL, NULL)) {
        unlocked = g_strrstr (output, "Unlocked: No") == NULL;
        g_free(output);
    }
    return unlocked;
}

gboolean is_encfs (const gchar *path) {
    gchar *encfs = g_strdup_printf ("%s/.encfs6.xml", path);
    gboolean initialized = g_file_test (encfs, G_FILE_TEST_EXISTS);
    g_free (encfs);
    return initialized;
}

gboolean is_cryfs (const gchar *path) {
    gchar *cryfs = g_strdup_printf ("%s/cryfs.config", path);
    gboolean initialized = g_file_test (cryfs, G_FILE_TEST_EXISTS);
    g_free (cryfs);
    return initialized;
}

gboolean is_fscrypt (const gchar *path) {
    gint wait_status;
    gchar *argv[] = { "fscrypt", "status", (gchar *) path, NULL };
    gboolean initialized = FALSE;
    if (g_spawn_sync (NULL, argv, NULL, G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_STDOUT_TO_DEV_NULL,
                      NULL, NULL, NULL, NULL, &wait_status, NULL)) {
        initialized = g_spawn_check_wait_status (wait_status, NULL);
    }
    return initialized;
}

gboolean is_empty (GFile *folder) {
    gboolean empty = TRUE;
    GFileEnumerator *children = g_file_enumerate_children (folder, G_FILE_ATTRIBUTE_STANDARD_NAME, G_FILE_QUERY_INFO_NONE, NULL, NULL);
    GFileInfo *child = g_file_enumerator_next_file(children, NULL, NULL);
    if (child != NULL) {
        empty = FALSE;
        g_object_unref (child);
    }
    g_object_unref (children);
    return empty;
}

gboolean is_fscrypt_available() {
    gboolean found = FALSE;
    gchar *path = g_find_program_in_path ("fscrypt");
    if (path != NULL) {
        g_free (path);
        found = TRUE;
    }
    return found;
}

gboolean is_encfs_available() {
    gboolean found = FALSE;
    gchar *path = g_find_program_in_path ("encfs");
    if (path != NULL) {
        g_free (path);
        found = TRUE;
    }
    return found;
}

gboolean is_cryfs_available() {
    gboolean found = FALSE;
    gchar *path = g_find_program_in_path ("cryfs");
    if (path != NULL) {
        g_free (path);
        found = TRUE;
    }
    return found;
}



//GUI
static void do_ask_password (const gchar *path, FolderStatus status) {
    GtkWidget *window = gtk_window_new_compat();
    gtk_window_set_title (GTK_WINDOW(window), "Thunar-Encryption-Plugin");
    gtk_window_set_icon_name (GTK_WINDOW(window), "encrypted-folder");
    gtk_window_set_default_size (GTK_WINDOW(window), 300, 0);
    gtk_window_set_resizable (GTK_WINDOW(window), FALSE);
    GtkWidget *vbox = gtk_box_new (GTK_ORIENTATION_VERTICAL, 8);
    gtk_window_set_child (GTK_WINDOW(window), vbox);
    gtk_widget_set_margin_top (vbox, 8);

    gtk_box_append (GTK_BOX(vbox), gtk_label_new (gettext ("Enter password for:")));
    gtk_box_append (GTK_BOX(vbox), gtk_label_new (path));
    GtkWidget *entry = gtk_password_entry_new();
    gtk_password_entry_set_show_peek_icon ((GtkPasswordEntry *) entry, TRUE); //GTK_PASSWORD_ENTRY macro not defined
    gtk_box_append (GTK_BOX(vbox), entry);

    GtkWidget *hbox = gtk_box_new (GTK_ORIENTATION_HORIZONTAL, 8);
    GtkWidget *cancel_button = gtk_button_new_with_label (gettext ("Cancel"));
    g_signal_connect_swapped (cancel_button, "clicked", G_CALLBACK(gtk_window_destroy), window);
    gtk_box_append (GTK_BOX(hbox), cancel_button);
    GtkWidget *ok_button = gtk_button_new_with_label (gettext ("OK"));
    gtk_widget_add_css_class (ok_button, "suggested-action");
    g_signal_connect (ok_button, "clicked", G_CALLBACK (on_password_validate), entry);
    g_object_set_data_full (G_OBJECT (entry), "path", g_strdup (path), g_free);
    g_object_set_data_full (G_OBJECT (entry), "status", g_strdup_printf ("%u", status), g_free);
    g_signal_connect_swapped (ok_button, "clicked", G_CALLBACK(gtk_window_destroy), window);
    gtk_box_append (GTK_BOX(hbox), ok_button);
    gtk_box_set_homogeneous (GTK_BOX(hbox), TRUE);
    gtk_box_append (GTK_BOX(vbox), hbox);
    gtk_window_set_default_widget (GTK_WINDOW(window), ok_button);
    gtk_window_present_compat (GTK_WINDOW(window));
}

static void on_password_validate (GtkButton *button, GtkWidget *entry) {
    const gchar *path = g_object_get_data (G_OBJECT (entry), "path");
    FolderStatus status = (FolderStatus) g_ascii_strtoll (g_object_get_data (G_OBJECT (entry), "status"), NULL, 10);
    const gchar *password = gtk_editable_get_text (GTK_EDITABLE(entry));
    if (status == FSCRYPT_FOLDER) {
        do_unlock_folder_fscrypt (path, password);
    } else if (status == ENCFS_FOLDER) {
        do_unlock_folder_fuse (path, password, FALSE);
    } else if (status == CRYFS_FOLDER) {
        do_unlock_folder_fuse (path, password, TRUE);
    } else if (status == DO_ENCRYPT_FS) {
        do_encrypt_folder_fscrypt (path, password);
    } else if (status == DO_ENCRYPT_FUSE) {
        do_encrypt_folder_fuse (path, password, is_cryfs_available());
    }
}

static void do_notify_open (const gchar *path, gboolean success) {
    gchar *message;
    gchar *icon;
    if (success) {
        message = g_strdup_printf (gettext ("Folder '%s' opened."), path);
        icon = "--icon=dialog-information";
    } else {
        message = g_strdup_printf (gettext ("Password error for '%s'."), path);
        icon = "--icon=dialog-error";
    }
    gchar *argv[] = { "notify-send", "-t", "5000", icon, message, NULL };
    g_spawn_sync (NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL, NULL, NULL);
    g_free (message);
}

static void do_notify_close (const gchar *path, gboolean success) {
    gchar *message;
    gchar *icon;
    if (success) {
        message = g_strdup_printf (gettext ("Folder '%s' closed."), path);
        icon = "--icon=dialog-information";
    } else {
        message = g_strdup_printf (gettext ("Folder '%s' is still opened."), path);
        icon = "--icon=dialog-error";
    }
    gchar *argv[] = { "notify-send", "-t", "5000", icon, message, NULL };
    g_spawn_sync (NULL, argv, NULL, G_SPAWN_SEARCH_PATH, NULL, NULL, NULL, NULL, NULL, NULL);
    g_free (message);
}



//FSCrypt functions
void do_unlock_folder_fscrypt(const gchar *path, const gchar *password) {
    GList *protectors = get_fscrypt_protectors (path);
    gchar *password_lf_ctrl_c = g_strdup_printf ("%s\n\x03", password);

    for (GList *current = protectors; current != NULL; current = current->next) {
        gchar *with_protector = g_strdup_printf ("--unlock-with=%s", (gchar *) current->data);
        gchar *argv[] = { "fscrypt", "unlock", (gchar *) path, with_protector, NULL };
        spawn_with_input_pipe_and_path (argv, NULL, password_lf_ctrl_c, path);
        g_free (with_protector);
    }

    g_free (password_lf_ctrl_c);
    g_list_free_full (protectors, g_free);
}

void do_encrypt_folder_fscrypt (const gchar *path, const gchar *password) {
    gchar *password_lf = g_strdup_printf ("%s\n", password);
    gchar *name_option = g_strdup_printf ("--name=%s", path);
    for (gchar *ptr = name_option + 7; *ptr != '\0'; ptr++) *ptr = *ptr == '/' ? '_' : *ptr;
    gchar *argv[] = { "fscrypt", "encrypt", (gchar *) path, name_option, "--no-recovery", "--quiet", NULL };
    spawn_with_input_pipe_and_path (argv, NULL, password_lf, path);
    g_free (name_option);
    g_free (password_lf);
}

static GList* get_fscrypt_protectors (const gchar* path) {
    GList* protectors = NULL;
    gchar* output = NULL;
    gchar *argv[] = { "fscrypt", "status", (gchar *) path, "--verbose", NULL };
    if (g_spawn_sync (NULL, argv, NULL, G_SPAWN_SEARCH_PATH | G_SPAWN_STDERR_TO_DEV_NULL, NULL, NULL, &output, NULL, NULL, NULL)) {
        gchar** lines = g_strsplit (output, "\n", -1);
        for (guint i=0; lines[i] != NULL; i++) {
            gchar *idx = g_strrstr (lines[i], "/protectors/"); //... "/home/.fscrypt/protectors/0123abcd4567ef89"
            if (idx != NULL) {
                gchar *id = g_strrstr (lines[i], "/"); id++; //from the last '/'
                id[strlen (id) - 1] = '\0'; //remove trailing '"'
                gchar *path = g_strrstr (lines[i], "\""); path++;
                idx -= 9; *idx = '\0'; //strlen("/.fscrypt")
                protectors = g_list_prepend (protectors, g_strdup_printf ("%s:%s", path, id)); ////-> /home:0123abcd4567ef89"
            }
        }
        g_strfreev (lines);
        g_free (output);
    }
    return protectors;
}



//FUSE (EncFs / CryFS) related functions
void do_unlock_folder_fuse(const gchar *path, const gchar *password, gboolean use_cryfs) {
    gchar *mount = get_mount_path (path);
    gchar *password_lf = g_strdup_printf ("%s\n", password);
    gchar *argv_cryfs[] = { "cryfs", (gchar *) path, mount, NULL };
    gchar *argv_encfs[] = { "encfs", "--stdinpass", (gchar *) path, mount, NULL };
    gchar *envp_cryfs[] = { "CRYFS_FRONTEND=noninteractive", NULL };
    create_mount_directory (path);
    spawn_with_input_pipe_and_path (use_cryfs ? argv_cryfs : argv_encfs, use_cryfs ? envp_cryfs : NULL, password_lf, path);
    g_free (password_lf);
    g_free (mount);
}

void do_encrypt_folder_fuse (const gchar *path, const gchar *password, gboolean use_cryfs) {
    gchar *mount = get_mount_path (path);
    gchar *input_string = use_cryfs ? g_strdup_printf("%s\n", password) : g_strdup_printf("p\n%s\n%s\n", password, password);
    gchar *argv_cryfs[] = { "cryfs", (gchar *) path, mount, NULL };
    gchar *argv_encfs[] = { "encfs", "--stdinpass", (gchar *) path, mount, NULL };
    gchar *envp_cryfs[] = { "CRYFS_FRONTEND=noninteractive", NULL };
    //g_setenv ("CRYFS_FRONTEND", "noninteractive", TRUE);
    create_mount_directory (path);
    spawn_with_input_pipe_and_path (use_cryfs ? argv_cryfs : argv_encfs, use_cryfs ? envp_cryfs : NULL, input_string, path);
    g_free (input_string);
    g_free (mount);
}

gchar * get_mount_path (const gchar *path) {
    GFile *file = g_file_new_for_path (path);
    gchar *basename = g_file_get_basename (file);
    g_object_unref (file);
    gchar *mount_path = g_strdup_printf ("%s/%s/%s", g_getenv("HOME"), FUSE_MOUNT_PATH, basename);
    g_free (basename);
    return mount_path;
}

void create_mount_directory (const char *path) {
    gchar *encrypted_root = g_strdup_printf ("%s/%s", g_getenv("HOME"), FUSE_MOUNT_PATH);
    gchar *mount_path = get_mount_path (path);
    if (!g_file_test (encrypted_root, G_FILE_TEST_IS_DIR)) {
       g_mkdir (encrypted_root, 0700);
    }
    if (!g_file_test (mount_path, G_FILE_TEST_IS_DIR)) {
       g_mkdir (mount_path, 0700);
    }
    g_free (mount_path);
    g_free (encrypted_root);
}

//FSCrypt, CryFS & EncFS commons
static void open_child_watch_cb (GPid pid, gint status, gpointer user_data) {
    do_notify_open ((gchar *) user_data, g_spawn_check_wait_status (status, NULL));
    g_free (user_data);
}

static void spawn_with_input_pipe_and_path (gchar **argv, gchar **envp, const gchar *input_string, const gchar *path) {
    gint child_stdin; GPid child_pid;
    if (g_spawn_async_with_pipes (NULL, argv, envp, G_SPAWN_SEARCH_PATH | G_SPAWN_DO_NOT_REAP_CHILD | G_SPAWN_STDERR_TO_DEV_NULL | G_SPAWN_STDOUT_TO_DEV_NULL,
                                  NULL, NULL, &child_pid, &child_stdin, NULL, NULL, NULL)) {
        GOutputStream *stdin_stream = g_unix_output_stream_new (child_stdin, TRUE);
        g_output_stream_write_all (stdin_stream, input_string, strlen (input_string), NULL, NULL, NULL);
        g_output_stream_close(stdin_stream, NULL, NULL);
        g_object_unref(stdin_stream);
        g_child_watch_add (child_pid, open_child_watch_cb, g_strdup (path)); //g_object_set_data_full -> g_free(path)
    }
}
